@@ -9,12 +9,20 @@ public:
   typedef typename Constants::Sizes Sizes;
 
   struct Options {
-    Options() : encode_srcwin_maxsz(1<<20), 
-		block_size(Constants::BLOCK_SIZE),
-		size_known(false) { }
-    size_t encode_srcwin_maxsz;
+    Options()
+      : encode_srcwin_maxsz(1<<20),
+	block_size(Constants::BLOCK_SIZE),
+	window_size(Constants::WINDOW_SIZE),
+	size_known(false),
+	iopt_size(XD3_DEFAULT_IOPT_SIZE),
+	smatch_cfg(XD3_SMATCH_DEFAULT) { }
+
+    xoff_t encode_srcwin_maxsz;
     size_t block_size;
+    xoff_t window_size;
     bool size_known;
+    usize_t iopt_size;
+    xd3_smatch_cfg smatch_cfg;
   };
 
 #include "segment.h"
@@ -26,7 +34,7 @@ public:
   void InMemoryEncodeDecode(const FileSpec &source_file,
 			    const FileSpec &target_file,
 			    Block *coded_data,
-			    const Options &options = Options()) {
+			    const Options &options) {
     xd3_stream encode_stream;
     xd3_config encode_config;
     xd3_source encode_source;
@@ -50,17 +58,9 @@ public:
     xd3_init_config(&encode_config, XD3_ADLER32);
     xd3_init_config(&decode_config, XD3_ADLER32);
 
-    encode_config.winsize = Constants::WINDOW_SIZE;
-
-    // TODO! the smatcher setup isn't working,
-  //   if (options.large_cksum_step) {
-  //     encode_config.smatch_cfg = XD3_SMATCH_SOFT;
-  //     encode_config.smatcher_soft.large_step = options.large_cksum_step;
-  //   }
-  //   if (options.large_cksum_size) {
-  //     encode_config.smatch_cfg = XD3_SMATCH_SOFT;
-  //     encode_config.smatcher_soft.large_look = options.large_cksum_size;
-  //   }
+    encode_config.winsize = options.window_size;
+    encode_config.iopt_size = options.iopt_size;
+    encode_config.smatch_cfg = options.smatch_cfg;
 
     CHECK_EQ(0, xd3_config_stream (&encode_stream, &encode_config));
     CHECK_EQ(0, xd3_config_stream (&decode_stream, &decode_config));
@@ -71,30 +71,30 @@ public:
     encode_source.max_winsize = options.encode_srcwin_maxsz;
     decode_source.max_winsize = options.encode_srcwin_maxsz;
 
-    if (!options.size_known) 
+    if (!options.size_known)
       {
 	xd3_set_source (&encode_stream, &encode_source);
 	xd3_set_source (&decode_stream, &decode_source);
       }
-    else 
+    else
       {
-	xd3_set_source_and_size (&encode_stream, &encode_source, 
+	xd3_set_source_and_size (&encode_stream, &encode_source,
 				 source_file.Size());
 	xd3_set_source_and_size (&decode_stream, &decode_source,
 				 source_file.Size());
       }
 
     BlockIterator source_iterator(source_file, options.block_size);
-    BlockIterator target_iterator(target_file, Constants::READ_SIZE);
+    BlockIterator target_iterator(target_file, Constants::WINDOW_SIZE);
     Block encode_source_block, decode_source_block;
     Block decoded_block, target_block;
     bool encoding = true;
     bool done = false;
     bool done_after_input = false;
 
-    IF_DEBUG1 (XPR(NTR "source %"Q"u[%"Q"u] target %"Q"u[%lu] winsize %lu\n",
+    IF_DEBUG1 (XPR(NTR "source %"Q"u[%"Q"u] target %"Q"u winsize %"Z"u\n",
 		  source_file.Size(), options.block_size,
-		  target_file.Size(), Constants::READ_SIZE,
+		  target_file.Size(),
 		  Constants::WINDOW_SIZE));
 
     while (!done) {
@@ -102,7 +102,8 @@ public:
 
       xoff_t blks = target_iterator.Blocks();
 
-      IF_DEBUG2(XPR(NTR "target in %s: %llu..%llu %"Q"u(%"Q"u) verified %"Q"u\n",
+      IF_DEBUG2(XPR(NTR "target in %s: %"Q"u..%"Q"u %"Q"u(%"Q"u) "
+		    "verified %"Q"u\n",
 		   encoding ? "encoding" : "decoding",
 		   target_iterator.Offset(),
 		   target_iterator.Offset() + target_block.Size(),
@@ -152,11 +153,11 @@ public:
 	xd3_source *src = (encoding ? &encode_source : &decode_source);
 	Block *block = (encoding ? &encode_source_block : &decode_source_block);
 	if (encoding) {
-	  IF_DEBUG1(XPR(NTR "[srcblock] %"Q"u last srcpos %"Q"u "
-		       "encodepos %"Q"u\n",
-		       encode_source.getblkno,
-		       encode_stream.match_last_srcpos,
-		       encode_stream.input_position + encode_stream.total_in));
+	  IF_DEBUG2(XPR(NTR "[srcblock] %"Q"u last srcpos %"Q"u "
+			"encodepos %"Q"u\n",
+			encode_source.getblkno,
+			encode_stream.match_last_srcpos,
+			encode_stream.input_position + encode_stream.total_in));
 	}
 
 	source_iterator.SetBlock(src->getblkno);
@@ -229,10 +230,10 @@ public:
 			ExtFile *coded_data,
 			const Options &options) {
     vector<const char*> ecmd;
-    char buf[16];
-    snprintf(buf, sizeof(buf), "-B%"Q"u", options.encode_srcwin_maxsz);
+    char bbuf[16];
+    snprintf(bbuf, sizeof(bbuf), "-B%"Q"u", options.encode_srcwin_maxsz);
     ecmd.push_back("xdelta3");
-    ecmd.push_back(buf);
+    ecmd.push_back(bbuf);
     ecmd.push_back("-s");
     ecmd.push_back(source_file.Name());
     ecmd.push_back(target_file.Name());
@@ -245,7 +246,7 @@ public:
     vector<const char*> dcmd;
     ExtFile recon_file;
     dcmd.push_back("xdelta3");
-    ecmd.push_back(buf);  
+    ecmd.push_back(bbuf);
     dcmd.push_back("-d");
     dcmd.push_back("-s");
     dcmd.push_back(source_file.Name());
@@ -256,32 +257,146 @@ public:
     CHECK_EQ(0, xd3_main_cmdline(dcmd.size() - 1,
 				 const_cast<char**>(&dcmd[0])));
 
-    CHECK_EQ(0, test_compare_files(recon_file.Name(), 
+    CHECK_EQ(0, test_compare_files(recon_file.Name(),
 				   target_file.Name()));
   }
 
- //////////////////////////////////////////////////////////////////////
+  // Similar to xd3_process_memory, with support for test Options.
+  // Exercises xd3_process_stream.
+  int TestProcessMemory (int            is_encode,
+			 int          (*func) (xd3_stream *),
+			 const uint8_t *input,
+			 usize_t        input_size,
+			 const uint8_t *source,
+			 usize_t        source_size,
+			 uint8_t       *output,
+			 usize_t       *output_size,
+			 usize_t        output_size_max,
+			 const Options &options) {
+    xd3_stream stream;
+    xd3_config config;
+    xd3_source src;
+    int ret;
 
- void TestRandomNumbers() {
-   MTRandom rand;
-   int rounds = 1<<20;
-   uint64_t usum = 0;
-   uint64_t esum = 0;
+    memset (& stream, 0, sizeof (stream));
+    memset (& config, 0, sizeof (config));
 
-   for (int i = 0; i < rounds; i++) {
-     usum += rand.Rand32();
-     esum += rand.ExpRand32(1024);
-   }
+    if (is_encode)
+      {
+	config.winsize = input_size;
+	config.iopt_size = options.iopt_size;
+	config.sprevsz = xd3_pow2_roundup (config.winsize);
+      }
 
-   double allowed_error = 0.01;
+    if ((ret = xd3_config_stream (&stream, &config)) != 0)
+      {
+	goto exit;
+      }
 
-   uint32_t umean = usum / rounds;
-   uint32_t emean = esum / rounds;
+    if (source != NULL)
+      {
+	memset (& src, 0, sizeof (src));
 
-   uint32_t uexpect = UINT32_MAX / 2;
-   uint32_t eexpect = 1024;
+	src.blksize = source_size;
+	src.onblk = source_size;
+	src.curblk = source;
+	src.curblkno = 0;
+	src.max_winsize = source_size;
 
-   if (umean < uexpect * (1.0 - allowed_error) ||
+	if ((ret = xd3_set_source_and_size (&stream, &src, source_size)) != 0)
+	  {
+	    goto exit;
+	  }
+      }
+
+    if ((ret = xd3_process_stream (is_encode,
+				   & stream,
+				   func, 1,
+				   input, input_size,
+				   output,
+				   output_size,
+				   output_size_max)) != 0)
+      {
+	goto exit;
+      }
+
+  exit:
+    if (ret != 0)
+      {
+	IF_DEBUG2 (DP(RINT "test_process_memory: %d: %s\n", ret, stream.msg));
+      }
+    xd3_free_stream(&stream);
+    return ret;
+  }
+
+  void EncodeDecodeAPI(const FileSpec &spec0, const FileSpec &spec1, 
+		       Block *delta, const Options &options) {
+    Block from;
+    Block to;
+    spec0.Get(&from, 0, spec0.Size());
+    spec1.Get(&to, 0, spec1.Size());
+
+    delta->SetSize(to.Size() * 1.5);
+    usize_t out_size;
+    int enc_ret = TestProcessMemory(true,
+				    &xd3_encode_input,
+				    to.Data(),
+				    to.Size(),
+				    from.Data(),
+				    from.Size(),
+				    delta->Data(),
+				    &out_size,
+				    delta->Size(),
+				    options);
+    CHECK_EQ(0, enc_ret);
+    delta->SetSize(out_size);
+
+    Block recon;
+    recon.SetSize(to.Size());
+    usize_t recon_size;
+    int dec_ret = xd3_decode_memory(delta->Data(),
+				    delta->Size(),
+				    from.Data(),
+				    from.Size(),
+				    recon.Data(),
+				    &recon_size,
+				    recon.Size(),
+				    0);
+    CHECK_EQ(0, dec_ret);
+    CHECK_EQ(0, CmpDifferentBlockBytes(to, recon));
+  }
+
+//////////////////////////////////////////////////////////////////////
+
+void TestPrintf() {
+  char buf[64];
+  xoff_t x = XOFF_T_MAX;
+  snprintf_func (buf, sizeof(buf), "%"Q"u", x);
+  const char *expect = XD3_USE_LARGEFILE64 ?
+    "18446744073709551615" : "4294967295";
+  XD3_ASSERT(strcmp (buf, expect) == 0);
+}
+
+void TestRandomNumbers() {
+  MTRandom rand;
+  int rounds = 1<<20;
+  uint64_t usum = 0;
+  uint64_t esum = 0;
+
+  for (int i = 0; i < rounds; i++) {
+    usum += rand.Rand32();
+    esum += rand.ExpRand32(1024);
+  }
+
+  double allowed_error = 0.01;
+
+  uint32_t umean = usum / rounds;
+  uint32_t emean = esum / rounds;
+
+  uint32_t uexpect = UINT32_MAX / 2;
+  uint32_t eexpect = 1024;
+
+  if (umean < uexpect * (1.0 - allowed_error) ||
       umean > uexpect * (1.0 + allowed_error)) {
     XPR(NT "uniform mean error: %u != %u\n", umean, uexpect);
     abort();
@@ -370,7 +485,7 @@ void TestFirstByte() {
     }
     spec0.GenerateFixedSize(size);
     spec0.ModifyTo(Modify1stByte(), &spec1);
-    InMemoryEncodeDecode(spec0, spec1, NULL);
+    InMemoryEncodeDecode(spec0, spec1, NULL, Options());
   }
 }
 
@@ -406,7 +521,7 @@ void TestModifyMutator() {
     // pass.
     CHECK_GE(diff, test_cases[i].size - (2 * test_cases[i].size / 256));
 
-    InMemoryEncodeDecode(spec0, spec1, NULL);
+    InMemoryEncodeDecode(spec0, spec1, NULL, Options());
   }
 }
 
@@ -439,7 +554,7 @@ void TestAddMutator() {
     CHECK_EQ(spec0.Size() + test_cases[i].size, spec1.Size());
 
     Block coded;
-    InMemoryEncodeDecode(spec0, spec1, &coded);
+    InMemoryEncodeDecode(spec0, spec1, &coded, Options());
 
     Delta delta(coded);
     CHECK_EQ(test_cases[i].expected_adds,
@@ -470,13 +585,13 @@ void TestDeleteMutator() {
 
   for (size_t i = 0; i < SIZEOF_ARRAY(test_cases); i++) {
     ChangeList cl1;
-    cl1.push_back(Change(Change::DELETE, test_cases[i].size,
+    cl1.push_back(Change(Change::DELRANGE, test_cases[i].size,
 			 test_cases[i].addr));
     spec0.ModifyTo(ChangeListMutator(cl1), &spec1);
     CHECK_EQ(spec0.Size() - test_cases[i].size, spec1.Size());
 
     Block coded;
-    InMemoryEncodeDecode(spec0, spec1, &coded);
+    InMemoryEncodeDecode(spec0, spec1, &coded, Options());
 
     Delta delta(coded);
     CHECK_EQ(0, delta.AddedBytes());
@@ -511,7 +626,7 @@ void TestCopyMutator() {
     CHECK_EQ(spec0.Size() + test_cases[i].size, spec1.Size());
 
     Block coded;
-    InMemoryEncodeDecode(spec0, spec1, &coded);
+    InMemoryEncodeDecode(spec0, spec1, &coded, Options());
 
     Delta delta(coded);
     CHECK_EQ(0, delta.AddedBytes());
@@ -553,7 +668,7 @@ void TestMoveMutator() {
     CHECK_EQ(spec0.Size(), spec1.Size());
 
     Block coded;
-    InMemoryEncodeDecode(spec0, spec1, &coded);
+    InMemoryEncodeDecode(spec0, spec1, &coded, Options());
 
     Delta delta(coded);
     CHECK_EQ(0, delta.AddedBytes());
@@ -596,7 +711,7 @@ void TestOverwriteMutator() {
 
 // Note: this test is written to expose a problem, but the problem was
 // only exposed with BLOCK_SIZE = 128.
-void TestNonBlockingProgress() {
+void TestNonBlocking() {
   MTRandom rand;
   FileSpec spec0(&rand);
   FileSpec spec1(&rand);
@@ -635,7 +750,7 @@ void TestNonBlockingProgress() {
 
   spec0.ModifyTo(ChangeListMutator(ctl), &spec2);
 
-  InMemoryEncodeDecode(spec1, spec2, NULL);
+  InMemoryEncodeDecode(spec1, spec2, NULL, Options());
 }
 
 void TestEmptyInMemory() {
@@ -647,7 +762,7 @@ void TestEmptyInMemory() {
   spec0.GenerateFixedSize(0);
   spec1.GenerateFixedSize(0);
 
-  InMemoryEncodeDecode(spec0, spec1, &block);
+  InMemoryEncodeDecode(spec0, spec1, &block, Options());
 
   Delta delta(block);
   CHECK_LT(0, block.Size());
@@ -663,94 +778,224 @@ void TestBlockInMemory() {
   spec0.GenerateFixedSize(Constants::BLOCK_SIZE);
   spec1.GenerateFixedSize(Constants::BLOCK_SIZE);
 
-  InMemoryEncodeDecode(spec0, spec1, &block);
+  InMemoryEncodeDecode(spec0, spec1, &block, Options());
 
   Delta delta(block);
   CHECK_EQ(spec1.Blocks(Constants::WINDOW_SIZE), delta.Windows());
 }
 
-void TestHalfBlockCopy() {
+void TestSmallStride() {
   MTRandom rand;
   FileSpec spec0(&rand);
-  FileSpec spec1(&rand);
+  usize_t size = Constants::BLOCK_SIZE * 4;
+  spec0.GenerateFixedSize(size);
 
-  spec0.GenerateFixedSize(Constants::BLOCK_SIZE * 4);
+  // Note: Not very good performance due to hash collisions, note 3x
+  // multiplier below.
+  for (int s = 15; s < 101; s++) {
+    usize_t changes = 0;
+    ChangeList cl;
+    for (usize_t j = s; j < size; j += s, ++changes)
+      {
+	cl.push_back(Change(Change::MODIFY, 1, j));
+      }
 
-  // Create a half-block copy, 2.5 blocks apart, from the second half
-  // of the source version to the first half of the target version.
-  //       0             1             2             3
-  // spec0 [bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb][ccccc][bbbbb]
-  // spec1 [aaaaa][ccccc][aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa]
-  ChangeList cl1;
-  cl1.push_back(Change(Change::MODIFY,
-		       Constants::BLOCK_SIZE / 2,  // size
-		       0));
-  cl1.push_back(Change(Change::COPYOVER,
-		       Constants::BLOCK_SIZE / 2,  // size
-		       Constants::BLOCK_SIZE * 3,  // offset
-		       Constants::BLOCK_SIZE / 2));
-  cl1.push_back(Change(Change::MODIFY,
-		       Constants::BLOCK_SIZE * 3,
-		       Constants::BLOCK_SIZE));
-  spec0.ModifyTo(ChangeListMutator(cl1), &spec1);
+    FileSpec spec1(&rand);
+    spec0.ModifyTo(ChangeListMutator(cl), &spec1);
 
-  const int onecopy_adds = 
-    4 * Constants::BLOCK_SIZE - Constants::BLOCK_SIZE / 2;
-  const int nocopy_adds = 4 * Constants::BLOCK_SIZE;
+    Options options;
+    options.encode_srcwin_maxsz = size;
+    options.iopt_size = 128;
+    options.smatch_cfg = XD3_SMATCH_SLOW;
+    options.size_known = false;
 
-  // Note the case b=4 is contrived: the caller should use a single block 
-  // containing the entire source, if possible.
-  for (int b = 1; b <= 4; b++)
+    Block block;
+    InMemoryEncodeDecode(spec0, spec1, &block, options);
+    Delta delta(block);
+
+    IF_DEBUG1(DP(RINT "[stride=%d] changes=%u adds=%"Q"u\n",
+		 s, changes, delta.AddedBytes()));
+    double allowance = Constants::BLOCK_SIZE < 8192 || s < 30 ? 3.0 : 1.1;
+    CHECK_GE(allowance * changes, (double)delta.AddedBytes());
+  }
+}
+
+void TestCopyWindow() {
+  // Construct an input that has many copies, to fill the IOPT buffer
+  // and force a source window decision.  "srclen" may be set to a
+  // value that goes beyond the end-of-source.
+  const int clen = 16;
+  const int size = 4096;
+  const int nmov = size / clen;
+  const int iters = 16;
+  uint32_t added_01 = 0;
+  uint32_t added_10 = 0;
+  for (int i = 1; i <= iters; i++) {
+    MTRandom rand(MTRandom::TEST_SEED1 * i);
+    FileSpec spec0(&rand);
+    ChangeList cl;
+
+    spec0.GenerateFixedSize(size);
+
+    for (int j = 0; j < nmov; j += 2)
+      {
+	cl.push_back(Change(Change::MOVE,
+			    clen, (j + 1) * clen, j * clen));
+      }
+
+    FileSpec spec1(&rand);
+    spec0.ModifyTo(ChangeListMutator(cl), &spec1);
+
+    Options options;
+    options.encode_srcwin_maxsz = size;
+    options.iopt_size = 128;
+    options.smatch_cfg = XD3_SMATCH_SLOW;
+
+    Block block1;
+    InMemoryEncodeDecode(spec0, spec1, &block1, options);
+    Delta delta1(block1);
+    // Allow one missed window (e.g., hash collisions)
+    added_01 += delta1.AddedBytes();
+
+    Block block2;
+    InMemoryEncodeDecode(spec1, spec0, &block2, options);
+    Delta delta2(block2);
+    // Allow one missed window (e.g., hash collisions)
+    added_10 += delta2.AddedBytes();
+
+    Block block3;
+    Block block4;
+    EncodeDecodeAPI(spec0, spec1, &block3, options);
+    EncodeDecodeAPI(spec1, spec0, &block4, options);
+  }
+  // Average less than 0.5 misses (of length clen) per iteration.
+  CHECK_GE(clen * iters / 2, added_01);
+  CHECK_GE(clen * iters / 2, added_10);
+}
+
+void TestCopyFromEnd() {
+  // Copies from the end of the source buffer, which reach a block
+  // boundary end-of-file.
+  const int size = 4096;
+  const int clen = 16;
+  const int nmov = (size / 2) / clen;
+  const int iters = 16;
+  uint32_t added_01 = 0;
+  uint32_t added_10 = 0;
+  for (int i = 1; i <= iters; i++) {
+    MTRandom rand(MTRandom::TEST_SEED1 * i);
+    FileSpec spec0(&rand);
+    ChangeList cl;
+
+    spec0.GenerateFixedSize(size);
+
+    cl.push_back(Change(Change::MODIFY, 2012, 2048));
+
+    for (int j = 0; j < nmov; j += 2)
+      {
+	cl.push_back(Change(Change::MOVE,
+			    clen, (j + 1) * clen, j * clen));
+      }
+
+    cl.push_back(Change(Change::COPYOVER, 28, 4068, 3000));
+    cl.push_back(Change(Change::COPYOVER, 30, 4066, 3100));
+    cl.push_back(Change(Change::COPYOVER, 32, 4064, 3200));
+
+    FileSpec spec1(&rand);
+    spec0.ModifyTo(ChangeListMutator(cl), &spec1);
+
+    Options options;
+    options.encode_srcwin_maxsz = size;
+    options.iopt_size = 128;
+    options.smatch_cfg = XD3_SMATCH_SLOW;
+
+    Block block1;
+    InMemoryEncodeDecode(spec0, spec1, &block1, options);
+    Delta delta1(block1);
+    added_01 += delta1.AddedBytes();
+
+    Block block2;
+    InMemoryEncodeDecode(spec1, spec0, &block2, options);
+    Delta delta2(block2);
+    added_10 += delta2.AddedBytes();
+
+    Block block3;
+    Block block4;
+    EncodeDecodeAPI(spec0, spec1, &block3, options);
+    EncodeDecodeAPI(spec1, spec0, &block4, options);
+  }
+  CHECK_GE(2000 * iters, added_01);
+  CHECK_LE(2000 * iters, added_10);
+}
+
+void TestHalfBlockCopy() {
+  // Create a half-block copy, 7.5 blocks apart, in a pair of files:
+  //       0             1     ...     6             7
+  // spec0 [bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb][ccccc][bbbb_]
+  // spec1 [aaaaa][ccccc][aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa_]
+  // where stage=
+  // 0: the final block is full
+  //   a. (source)spec1->(target)spec0 copies block C: reads 8 source
+  //      blocks during target block 0.
+  //   b. (source)spec0->(target)spec1 does not copy block C b/c attempt
+  //      to read past EOF empties block 0 from (virtual) block cache
+  // 1: the final block is less than full.
+  //   a. (same) copies block C
+  //   b. (same) copies block C, unlike 0a, no attempt to read past EOF
+  //
+  // "virtual" above refers to XD3_TOOFARBACK, since there is no caching
+  // in the API, there is simply a promise not to request blocks that are
+  // beyond source->max_winsize from the last known source file position.
+  for (int stage = 0; stage < 2; stage++)
     {
+      IF_DEBUG1 (DP(RINT "half_block_copy stage %d\n", stage));
+
+      MTRandom rand;
+      FileSpec spec0(&rand);
+      FileSpec spec1(&rand);
+
+      spec0.GenerateFixedSize(Constants::BLOCK_SIZE * 8 - stage);
+
+      ChangeList cl1;
+      cl1.push_back(Change(Change::MODIFY,
+			   Constants::BLOCK_SIZE / 2,  // size
+			   0));
+      cl1.push_back(Change(Change::COPYOVER,
+			   Constants::BLOCK_SIZE / 2,  // size
+			   Constants::BLOCK_SIZE * 7,  // offset
+			   Constants::BLOCK_SIZE / 2));
+      cl1.push_back(Change(Change::MODIFY,
+			   Constants::BLOCK_SIZE * 7,
+			   Constants::BLOCK_SIZE - stage));
+      spec0.ModifyTo(ChangeListMutator(cl1), &spec1);
+
       Options options;
-      options.encode_srcwin_maxsz = Constants::BLOCK_SIZE * b;
+      options.encode_srcwin_maxsz = Constants::BLOCK_SIZE * 8;
 
       Block block0;
       Block block1;
       InMemoryEncodeDecode(spec0, spec1, &block0, options);
       InMemoryEncodeDecode(spec1, spec0, &block1, options);
+
       Delta delta0(block0);
       Delta delta1(block1);
 
-      // The first block never copies from the last source block, by
-      // design, because if the last source block is available when
-      // the first target block is ready, the caller is expected to
-      // use a single block.
-      CHECK_EQ(nocopy_adds, delta0.AddedBytes());
-      if (Constants::BLOCK_SIZE < 8192 || b > 2)
-      	{
-	  // For small-block inputs, the entire file is read into one
-	  // block (the min source window size is 16kB).
-	  // 
-	  // For large blocks, at least 3 blocks of source window are
-	  // needed.
-      	  CHECK_EQ(onecopy_adds, delta1.AddedBytes());
-      	}
+      const int yes =
+	Constants::BLOCK_SIZE * 8 - Constants::BLOCK_SIZE / 2;
+      const int no =
+	Constants::BLOCK_SIZE * 8 - Constants::BLOCK_SIZE / 2;
+
+      if (stage == 0)
+	{
+	  CHECK_EQ(yes, delta0.AddedBytes());
+	  CHECK_EQ(no, delta1.AddedBytes());
+	}
       else
-      	{
-	  // When there are fewer than 3 source blocks.
-      	  CHECK_EQ(nocopy_adds, delta1.AddedBytes());
-      	}
-      // XPR(NT "0=%zu 1=%zu\n", delta0.AddedBytes(), delta1.AddedBytes());
+	{
+	  CHECK_EQ(yes, delta0.AddedBytes());
+	  CHECK_EQ(yes, delta1.AddedBytes());
+	}
     }
-
-  Options options;
-  options.encode_srcwin_maxsz = Constants::BLOCK_SIZE * 4;
-  options.block_size = Constants::BLOCK_SIZE * 4;
-
-  // Test the whole-buffer case.
-  Block block0;
-  Block block1;
-  InMemoryEncodeDecode(spec0, spec1, &block0, options);
-  InMemoryEncodeDecode(spec1, spec0, &block1, options);
-  Delta delta0(block0);
-  Delta delta1(block1);
-  // This <= >= are only for blocksize = 512, which has irregular readsize.
-  CHECK_LE(onecopy_adds, delta0.AddedBytes());
-  CHECK_GE(onecopy_adds + 1, delta0.AddedBytes());
-
-  CHECK_EQ(onecopy_adds, delta1.AddedBytes());
-  // XPR(NT "0=%zu 1=%zu\n", delta0.AddedBytes(), delta1.AddedBytes());
 }
 
 void FourWayMergeTest(const FileSpec &spec0,
@@ -760,7 +1005,7 @@ void FourWayMergeTest(const FileSpec &spec0,
   TmpFile f0, f1, f2, f3;
   ExtFile d01, d12, d23;
   Options options;
-  options.encode_srcwin_maxsz = 
+  options.encode_srcwin_maxsz =
     std::max(spec0.Size(), options.encode_srcwin_maxsz);
 
   spec0.WriteTmpFile(&f0);
@@ -883,7 +1128,7 @@ void TestMergeCommand1() {
       }
 
       cl1.push_back(Change(Change::ADD, change1, add1_pos));
-      cl2.push_back(Change(Change::DELETE, change1, del2_pos));
+      cl2.push_back(Change(Change::DELRANGE, change1, del2_pos));
       cl3.push_back(Change(Change::MODIFY, change3, change3_pos));
 
       spec0.ModifyTo(ChangeListMutator(cl1), &spec1);
@@ -937,9 +1182,9 @@ void TestMergeCommand2() {
 
 	  ChangeList cl1, cl2, cl3;
 
-	  cl1.push_back(Change(Change::DELETE, size0 - size1, 0));
-	  cl2.push_back(Change(Change::DELETE, size0 - size2, 0));
-	  cl3.push_back(Change(Change::DELETE, size0 - size3, 0));
+	  cl1.push_back(Change(Change::DELRANGE, size0 - size1, 0));
+	  cl2.push_back(Change(Change::DELRANGE, size0 - size2, 0));
+	  cl3.push_back(Change(Change::DELRANGE, size0 - size3, 0));
 
 	  spec0.ModifyTo(ChangeListMutator(cl1), &spec1);
 	  spec0.ModifyTo(ChangeListMutator(cl2), &spec2);
@@ -953,6 +1198,44 @@ void TestMergeCommand2() {
   }
 }
 
+void TestLastFrontierBlock() {
+  // This test constructs an input that can expose
+  // https://github.com/jmacd/xdelta/issues/188
+  // when run through the command-line with source via a FIFO.
+  // That is not tested here, but the test stays.
+  if (Constants::WINDOW_SIZE < XD3_ALLOCSIZE)
+    {
+      return;
+    }
+
+  MTRandom rand;
+  FileSpec spec0(&rand);
+  FileSpec spec1(&rand);
+  const xoff_t size = XD3_ALLOCSIZE * 64;  // == XD3_MINSRCWINSZ * 2
+  const xoff_t edit = XD3_ALLOCSIZE;
+
+  Options options;
+  options.encode_srcwin_maxsz = XD3_MINSRCWINSZ;
+  options.block_size = XD3_ALLOCSIZE;
+  options.window_size = XD3_MINSRCWINSZ;
+  options.size_known = false;
+
+  spec0.GenerateFixedSize(size);
+
+  ChangeList cl;
+
+  // Modify the 0th byte in order to induce indexing of subsequent
+  // bytes, but allow copying most of the file to keep the test fast.
+  cl.push_back(Change(Change::MODIFY, 1, edit * 31));
+  cl.push_back(Change(Change::COPYOVER, edit, edit * 31, edit * 63));
+
+  spec0.ModifyTo(ChangeListMutator(cl), &spec1);
+
+  Block noblock;
+  InMemoryEncodeDecode(spec0, spec1, &noblock, options);
+  InMemoryEncodeDecode(spec1, spec0, &noblock, options);
+}
+
 };  // class Regtest<Constants>
 
 #define TEST(x) XPR(NTR #x "...\n"); regtest.x()
@@ -961,6 +1244,7 @@ void TestMergeCommand2() {
 template <class T>
 void UnitTest() {
   Regtest<T> regtest;
+  TEST(TestPrintf);
   TEST(TestRandomNumbers);
   TEST(TestRandomFile);
   TEST(TestFirstByte);
@@ -975,20 +1259,24 @@ void UnitTest() {
 // These are Xdelta tests.
 template <class T>
 void MainTest() {
-  XPR(NT "Blocksize %"Q"u readsize %"Q"u windowsize %"Q"u\n", 
-      T::BLOCK_SIZE, T::READ_SIZE, T::WINDOW_SIZE);
+  XPR(NT "Blocksize %"Q"u windowsize %"Z"u\n",
+      T::BLOCK_SIZE, T::WINDOW_SIZE);
   Regtest<T> regtest;
   TEST(TestEmptyInMemory);
   TEST(TestBlockInMemory);
-  TEST(TestNonBlockingProgress);
+  TEST(TestSmallStride);
+  TEST(TestCopyWindow);
+  TEST(TestCopyFromEnd);
+  TEST(TestNonBlocking);
   TEST(TestHalfBlockCopy);
+  TEST(TestLastFrontierBlock);
   TEST(TestMergeCommand1);
   TEST(TestMergeCommand2);
 }
 
 #undef TEST
 
-int main(int argc, char **argv) 
+int main(int argc, char **argv)
 {
   vector<const char*> mcmd;
   string pn;
@@ -1004,7 +1292,6 @@ int main(int argc, char **argv)
   UnitTest<SmallBlock>();
   MainTest<SmallBlock>();
   MainTest<MixedBlock>();
-  MainTest<PrimeBlock>();
   MainTest<OversizeBlock>();
   MainTest<LargeBlock>();
 
@@ -1013,3 +1300,4 @@ int main(int argc, char **argv)
 
   return 0;
 }
+
