@@ -456,7 +456,7 @@ djw_build_prefix (const djw_weight *freq, uint8_t *clen, usize_t asize, usize_t
       djw_heapen *h2 = heap_extract (heap, ents, --heap_last);
 
       ents[ents_size].freq   = h1->freq + h2->freq;
-      ents[ents_size].depth  = 1 + max (h1->depth, h2->depth);
+      ents[ents_size].depth  = 1 + xd3_max (h1->depth, h2->depth);
       ents[ents_size].parent = 0;
 
       h1->parent = h2->parent = ents_size;
@@ -525,7 +525,7 @@ djw_build_codes (usize_t *codes, const uint8_t *clen, usize_t asize, usize_t abs
 	  min_clen = clen[i];
 	}
 
-      max_clen = max (max_clen, (usize_t) clen[i]);
+      max_clen = xd3_max (max_clen, (usize_t) clen[i]);
     }
 
   XD3_ASSERT (max_clen <= abs_max);
@@ -869,7 +869,7 @@ xd3_encode_huff (xd3_stream   *stream,
   usize_t     groups, sector_size;
   bit_state   bstate = BIT_STATE_ENCODE_INIT;
   xd3_output *in;
-  int         output_bits;
+  usize_t     output_bits;
   usize_t     input_bits;
   usize_t     input_bytes;
   usize_t     initial_offset = output->next;
@@ -1456,7 +1456,7 @@ djw_decode_symbol (xd3_stream     *stream,
       if (*input == input_end)
 	{
 	  stream->msg = "secondary decoder end of input";
-	  return XD3_INTERNAL;
+	  return XD3_INVALID_INPUT;
 	}
 
       bstate->cur_byte = *(*input)++;
@@ -1479,7 +1479,7 @@ djw_decode_symbol (xd3_stream     *stream,
 
  corrupt:
   stream->msg = "secondary decoder invalid code";
-  return XD3_INTERNAL;
+  return XD3_INVALID_INPUT;
 }
 
 static int
@@ -1606,7 +1606,7 @@ djw_decode_1_2 (xd3_stream     *stream,
   if (rep != 0)
     {
       stream->msg = "secondary decoder invalid repeat code";
-      return XD3_INTERNAL;
+      return XD3_INVALID_INPUT;
     }
   
   return 0;
@@ -1654,7 +1654,7 @@ xd3_decode_huff (xd3_stream     *stream,
   if (output_bytes == 0)
     {
       stream->msg = "secondary decoder invalid input";
-      return XD3_INTERNAL;
+      return XD3_INVALID_INPUT;
     }
 
   /* Decode: number of groups */
@@ -1796,10 +1796,14 @@ xd3_decode_huff (xd3_stream     *stream,
 		gp_maxlen  = maxlen[gp];
 	      }
 
-	    XD3_ASSERT (output_end - output > 0);
+	    if (output_end < output)
+	      {
+		stream->msg = "secondary decoder invalid input";
+		return XD3_INVALID_INPUT;
+	      }
 	    
 	    /* Decode next sector. */
-	    n = min (sector_size, (usize_t) (output_end - output));
+	    n = xd3_min (sector_size, (usize_t) (output_end - output));
 
 	    do
 	      {
